var uuid = require('node-uuid');

function ProcessScanner(agent) {
  this.agent = agent;
  this.autoSnapshotDurationSeconds = 10;
  this.snapshotCount = 0;
  this.maxSnapshotsPerPeriod = 2;
  this.snapshotCountResetPeriodMS = 1 * 60 * 1000;
  this.snapshotsResetIntervalId = undefined;
  this.currentProcessSnapshot = undefined;
  this.lastProcessRequestId = undefined;
  this.btCalls = null;
}

exports.ProcessScanner = ProcessScanner;

ProcessScanner.prototype.init = function () {
  var self = this;

  var snapshotsPerPeriod =
    self.agent.opts.maxProcessSnapshotsPerPeriod;

  if (snapshotsPerPeriod !== undefined) {
    snapshotsPerPeriod = +snapshotsPerPeriod;
    if (snapshotsPerPeriod >= 0) {
      self.maxSnapshotsPerPeriod = snapshotsPerPeriod;
    }
  }

  self.agent.logger.log("ProcessScanner - maxSnapshotsPerPeriod: " + self.maxSnapshotsPerPeriod);

  var resetPeriod = +(self.agent.opts.processSnapshotCountResetPeriodSeconds);
  if (resetPeriod && (resetPeriod > 0)) {
    self.snapshotCountResetPeriodMS = (+resetPeriod) * 1000;
  }
  self.agent.logger.log("ProcessScanner - snapshotCountResetPeriodMS: " + self.snapshotCountResetPeriodMS);

  var snapshotDuration = +(self.agent.opts.autoSnapshotDurationSeconds);
  if (snapshotDuration && (snapshotDuration > 0) && (snapshotDuration < 300)) {
    self.autoSnapshotDurationSeconds = snapshotDuration;
  }
  self.agent.logger.log("ProcessScanner - autoSnapshotDurationSeconds: " + self.autoSnapshotDurationSeconds);

  self.agent.on('transactionStarted', function(transaction, req) {
    if (!self.capturingProcessSnapshot) return;
    var threadId = transaction.threadId;
    var btCallInfo = self.btCalls[threadId] || (self.btCalls[threadId] = {});
    btCallInfo.startTime = self.agent.system.hrtime();
    btCallInfo.registrationId = transaction.registrationId;
    transaction.addProcessSnapshotGUID(self.currentProcessSnapshot.guid);
  });

  self.agent.on('transaction', function(transaction) {
    if (!self.capturingProcessSnapshot) return;

    transaction.addProcessSnapshotGUID(self.currentProcessSnapshot.guid);
    var threadId = transaction.threadId;
    var btCallInfo = self.btCalls[threadId] || (self.btCalls[threadId] = { startTime : undefined });
    btCallInfo.registrationId = btCallInfo.registrationId || transaction.registrationId;
    btCallInfo.endTime = self.agent.system.hrtime();
  });
}

Object.defineProperty(ProcessScanner.prototype, 'capturingProcessSnapshot', {
    get: function(){
        return !!this.currentProcessSnapshot;
    }
});

function ensureInterval(self) {
  if (self.snapshotsResetIntervalId)
    return;
  self.snapshotsResetIntervalId = self.agent.timers.setInterval(function () {
    self.snapshotCount = 0;
  }, self.snapshotCountResetPeriodMS);
}

// translate an undefined callback to a function that
// logs errors and sends the process snapshot to the proxy
// on success.
function ensureCallback(self, callback)
{
  return callback || function (err, processSnapshot) {
    if (err) {
      self.agent.logger.error(err);
      return;
    }
    self.agent.proxyTransport.sendProcessSnapshot(processSnapshot);
  };
}

ProcessScanner.prototype.doErrorCallback = function(msg, callback)
{
    if (!callback) {
        return;
    }

    var err = new Error(msg);
    process.nextTick(function() { callback(err, null); });
}

ProcessScanner.prototype.startAutoSnapshotIfPossible = function (callback) {
  var self = this;
  if (self.capturingProcessSnapshot) {
    var msg = 'Process snapshot already in progress, not starting another one!';
    self.agent.logger.log(msg);
    self.doErrorCallback(msg, callback);
    return;
  }

  if (self.snapshotCount >= self.maxSnapshotsPerPeriod) {
    var msg = 'Already did '
        + self.snapshotCount
        + ' process snapshots in '
        + self.snapshotCountResetPeriodMS
        + 'ms.'
    self.agent.logger.log(msg);
    self.doErrorCallback(msg, callback);
    return;
  }

  self.startSnapshot(self.autoSnapshotDurationSeconds, -1, ensureCallback(self, callback));
}

ProcessScanner.prototype.startManualSnapshot = function(processCallGraphReq, callback) {
  var self = this, requestID;

  requestID = +processCallGraphReq.snapshotRequestID;
  if(requestID <= self.lastSnapshotRequestID) {
    var msg = 'snapshotRequestID '
        + processCallGraphReq.snapshotRequestID
        + ' was already processed, ignoring.';
    self.agent.logger.error(msg);
    self.doErrorCallback(msg, callback);
    return;
  }

  self.lastSnapshotRequestID = requestID;

  if (self.capturingProcessSnapshot) {
    var msg = 'Process snapshot capture already in progress; snapshotRequestID '
        + processCallGraphReq.snapshotRequestID
        + ' ignored.'
    self.agent.logger.error(msg);
    self.doErrorCallback(msg, callback);
    return;
  }

  self.startSnapshot(processCallGraphReq.captureTime, requestID, ensureCallback(self, callback));
}

ProcessScanner.prototype.__populatePendingTransactions = function() {
  var self = this
  var startTime = self.agent.system.hrtime();
  var transactionsMap = self.agent.profiler.transactions;
  self.btCalls = {};
  for(var threadId in transactionsMap) {
      var btCallInfo = self.btCalls[threadId] || ( self.btCalls[threadId] = {} );
      btCallInfo.startTime = startTime;
      btCallInfo.registrationId =
        btCallInfo.registrationId || transactionsMap[threadId].registrationId;
  }
}

ProcessScanner.prototype.__btCallsToProto = function(btCalls) {
  var self = this;
  var result = [];
  var registrationIDToIndex = {};
  var now = self.agent.system.hrtime();
  for (var threadId in btCalls) {
    var btCallInfo = btCalls[threadId];
    if (btCallInfo.startTime === undefined)
      continue;
    if (!btCallInfo.registrationId)
      continue;
    if (btCallInfo.endTime === undefined)
      btCallInfo.endTime = now;

    var index = registrationIDToIndex[btCallInfo.registrationId];
    if (index === undefined) {
      index = result.length;
      registrationIDToIndex[btCallInfo.registrationId] = index;
    }

    var protoBTCallInfo =
      result[index] || ( result[index] = { btID: btCallInfo.registrationId, count: 0, totalTimeTakenMS: 0 } );

    protoBTCallInfo.count++;
    protoBTCallInfo.totalTimeTakenMS +=
      Math.max(0, (btCallInfo.endTime - btCallInfo.startTime) / 1000.0);
  }
  return result;
}

ProcessScanner.prototype.startSnapshot = function (durationSeconds,
                                                   requestID,
                                                   callback) {
  var self = this;

  if (self.capturingProcessSnapshot) {
    var msg = "Process snapshot already started, not starting another.";
    self.doErrorCallback(msg, callback);
    return;
  }

  self.__populatePendingTransactions();

  self.currentProcessSnapshot = {
      snapshotRequestID: requestID,
      guid : uuid.v4(),
      timestamp : undefined,
      processCallgraph : undefined,
      processID : process.pid,
      btCalls : undefined
  };
  self.snapshotCount++;
  ensureInterval(self);
  process.nextTick(function() {
    try {
      self.agent.cpuProfiler.startCpuProfiler(durationSeconds, function(err, processCallGraph) {
        try {
          if(err) {
            callback(err, null);
            return;
          }

          root = processCallGraph.callElements && processCallGraph.callElements[0];
          if (root && root.klass == "(global)" && root.method == "(root)") {
            processCallGraph.numOfRootElements = root.numChildren;
            processCallGraph.callElements.splice(0, 1);
          }

          var btCalls = self.__btCallsToProto(self.btCalls);

          var processSnapshot = self.currentProcessSnapshot;
          processSnapshot.timestamp = Date.now();
          processSnapshot.processCallGraph = processCallGraph;
          processSnapshot.btCalls = btCalls;

          self.lastSnapshotRequestID = requestID;
          callback(null, processSnapshot);
        } finally {
          self.btCalls = null;
          self.currentProcessSnapshot = undefined;
        }
      });
      self.agent.profiler.processSnapshotStarted(self.currentProcessSnapshot);
    }
    catch(err) {
      self.btCalls = null;
      self.currentProcessSnapshot = undefined;
      callback(err, null);
      return;
    }
  });
}
